#    This file is part of MDANSE_GUI.
#
#    MDANSE_GUI is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program.  If not, see <https://www.gnu.org/licenses/>.
#

from qtpy.QtWidgets import (
    QWidget,
    QVBoxLayout,
    QTableView,
    QPushButton,
    QHBoxLayout,
    QMessageBox,
)
from qtpy.QtCore import Slot, Signal

from MDANSE.MLogging import LOG

from MDANSE_GUI.Tabs.Models.PlottingContext import (
    PlottingContext,
    SingleDataset,
    plotting_column_labels,
)


class DataPlotter(QWidget):
    """This part of the interface will show the selection of datasets
    created by the user, and allow the creation of a plot."""

    error = Signal(str)
    data_for_plotting = Signal(object)
    data_for_new_plot = Signal(object)
    create_new_plot = Signal(str)
    create_new_text = Signal(str)

    def __init__(self, *args, unit_lookup=None, **kwargs):
        super().__init__(*args, **kwargs)

        self._unit_lookup = unit_lookup
        layout = QVBoxLayout(self)
        button_bar = QWidget(self)
        button_layout = QHBoxLayout(button_bar)
        self._selection_viewer = QTableView(self)
        layout.addWidget(self._selection_viewer)
        layout.addWidget(button_bar)
        buttons = [
            ("Plot Data", self.plot_data),
            ("Clear", self.clear),
            ("New Plot", self.new_plot),
            ("New Data View (Text)", self.new_text),
        ]
        for name, function in buttons:
            button = QPushButton(name, button_bar)
            button_layout.addWidget(button)
            if function is not None:
                button.clicked.connect(function)
        self._model = PlottingContext(
            unit_lookup=self._unit_lookup,
        )
        self._selection_viewer.setModel(self._model)

    @Slot(object)
    def add_dataset(self, dataset: SingleDataset):
        if not dataset._valid:
            return
        self._model.add_dataset(dataset)
        self._selection_viewer.resizeColumnsToContents()
        for col_num in range(4, 9):
            self._selection_viewer.hideColumn(col_num)

    @Slot()
    def new_plot(self):
        self.create_new_plot.emit("")
        group = self._settings.group("dialogs")
        try:
            show_it = group.get("new_plot")
        except KeyError:
            show_it = group.get_default("dialogs", "new_plot")
        if show_it != "False":
            plot_added_box = QMessageBox.information(
                self,
                "Plot created",
                "A new plot has been created in the next tab (called 'Plot Holder').\n"
                "Should this message be shown every time this happens?",
                QMessageBox.StandardButton.Yes | QMessageBox.StandardButton.No,
                QMessageBox.StandardButton.Yes,
            )
            if plot_added_box == QMessageBox.StandardButton.No:
                group = self._settings.group("dialogs")
                group.set("new_plot", "False")

    @Slot()
    def new_text(self):
        self.create_new_text.emit("Text view")
        group = self._settings.group("dialogs")
        try:
            show_it = group.get("new_text")
        except KeyError:
            show_it = group.get_default("dialogs", "new_text")
        if show_it != "False":
            plot_added_box = QMessageBox.information(
                self,
                "Plot created",
                "A new text view has been created in the next tab (called 'Plot Holder').\n"
                "Should this message be shown every time this happens?",
                QMessageBox.StandardButton.Yes | QMessageBox.StandardButton.No,
                QMessageBox.StandardButton.Yes,
            )
            if plot_added_box == QMessageBox.StandardButton.No:
                group = self._settings.group("dialogs")
                group.set("new_text", "False")

    @Slot()
    def plot_data(self):
        if len(self._model.datasets()) == 0:
            return
        self.data_for_plotting.emit(self._model)
        group = self._settings.group("dialogs")
        try:
            show_it = group.get("data_plotted")
        except KeyError:
            show_it = group.get_default("dialogs", "data_plotted")
        if show_it != "False":
            data_plotted_box = QMessageBox.information(
                self,
                "Datasets plotted!",
                "Your results have been plotted in the currently active plot in the next tab (called 'Plot Holder').\n"
                "Should this message be shown every time this happens?",
                QMessageBox.StandardButton.Yes | QMessageBox.StandardButton.No,
                QMessageBox.StandardButton.Yes,
            )
            if data_plotted_box == QMessageBox.StandardButton.No:
                group = self._settings.group("dialogs")
                group.set("data_plotted", "False")

    @Slot(object)
    def accept_data(self, data_set):
        LOG.info(f"Received {data_set}")
        dataset = SingleDataset(data_set[0], data_set[1])
        self.add_dataset(dataset)

    @Slot()
    def clear(self):
        if self._model is None:
            return
        self._model.clear()
        LOG.info(f"Cleared the plot")



import pytest
import tempfile
import os

import h5py
import numpy as np


from MDANSE_GUI.Session.StructuredSession import SettingsFile, SettingsGroup


@pytest.fixture(scope="module")
def settings_file():
    temp_name = tempfile.mktemp()
    path, name = os.path.split(temp_name)
    sf = SettingsFile(name, path)
    yield sf
    os.remove(temp_name + ".toml")


def test_loading(settings_file: "SettingsFile"):
    new_group = settings_file.group("blam")
    new_group.add("blob", "000", "useless test variable")
    settings_file.save_values()
    settings_file.load_from_file()
    loaded_group = settings_file.group("blam")
    loaded_value = loaded_group.get("blob")
    assert loaded_value == "000"


def test_reloading():
    temp_name = tempfile.mktemp()
    path, name = os.path.split(temp_name)
    settings_file = SettingsFile(name, path)
    new_group = settings_file.group("blam")
    new_group.add("blob", "000", "useless test variable")
    settings_file.save_values()
    settings_file2 = SettingsFile(name, path)
    settings_file2.load_from_file()
    loaded_group = settings_file2.group("blam")
    loaded_value = loaded_group.get("blob")
    assert loaded_value == "000"



#    This file is part of MDANSE_GUI.
#
#    MDANSE_GUI is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program.  If not, see <https://www.gnu.org/licenses/>.
#
from functools import partial
from qtpy.QtCore import Slot
from qtpy.QtWidgets import QWidget, QComboBox, QLabel

from MDANSE.MLogging import LOG

from MDANSE_GUI.Tabs.GeneralTab import GeneralTab
from MDANSE_GUI.Tabs.Layouts.MultiPanel import MultiPanel
from MDANSE_GUI.Session.LocalSession import LocalSession
from MDANSE_GUI.Tabs.Visualisers.Action import Action
from MDANSE_GUI.Tabs.Visualisers.TextInfo import TextInfo
from MDANSE_GUI.Tabs.Models.JobTree import JobTree
from MDANSE_GUI.Tabs.Views.ActionsTree import ActionsTree


job_tab_label = """This is the list of jobs
you can run using MDANSE.
Pick a job to see additional information.
Use the button to start a job,
or to save the inputs as a script.
"""


class JobTab(GeneralTab):
    """The tab for choosing and starting a new job."""

    def __init__(self, *args, **kwargs):
        self._needs_updating = False
        self.action = kwargs.pop("action")
        cmodel = kwargs.pop("combo_model", None)
        imodel = kwargs.pop("instrument_model", None)
        super().__init__(*args, **kwargs)
        self._current_trajectory = ""
        self._job_starter = None
        self._instrument_index = -1
        self._trajectory_combo = QComboBox()
        self._trajectory_combo.setEditable(False)
        self._trajectory_combo.currentIndexChanged.connect(self.set_current_trajectory)
        if cmodel is not None:
            self._trajectory_combo.setModel(cmodel)
        self._instrument_combo = QComboBox()
        self._instrument_combo.setEditable(False)
        self._instrument_combo.currentIndexChanged.connect(self.set_current_instrument)
        if imodel is not None:
            self._instrument_combo.setModel(imodel)
        self._core.add_widget(QLabel("Trajectory:"))
        self._core.add_widget(self._trajectory_combo)
        self._core.add_widget(QLabel("Instrument:"), upper=False)
        self._core.add_widget(self._instrument_combo, upper=False)
        self.action._parent_tab = self
        self._visualiser._parent_tab = self

    def set_job_starter(self, job_starter):
        self._job_starter = job_starter
        self.action.new_thread_objects.connect(self._job_starter.startProcess)
        self.action.run_and_load.connect(self._job_starter.startProcessAndLoad)

    def grouped_settings(self):
        results = super().grouped_settings()
        results += [
            [
                "Execution",
                {"auto-load": "True"},
                {
                    "auto-load": "Unless manually switched off, the GUI will try to load the job results when the job is finished."
                },
            ]
        ]
        return results

    @Slot(int)
    def set_current_trajectory(self, index: int) -> None:
        self._current_trajectory = self._trajectory_combo.currentText()

        traj_model = self._trajectory_combo.model()
        if traj_model.rowCount() < 1:
            # the combobox changed and there are no trajectories, they
            # were probably deleted lets clear the action widgets
            self.action.set_trajectory(path=None, trajectory=None)
            self.action.clear_panel()
            return

        node_number = traj_model.item(index, 0).data()
        LOG.info(
            f"Combo model: node_number {node_number} found in item {self._current_trajectory}"
        )
        # The combobox was changed we need to update the action
        # widgets with the new trajectory
        self.action.set_trajectory(
            path=None, trajectory=traj_model._nodes[node_number][0]
        )
        current_item = self._core.current_item()
        if current_item is not None:
            # we only update the widget if a job is selected from the
            # actions tree
            self.action.update_panel(current_item.text())

    @Slot(int)
    def set_current_instrument(self, index: int):
        LOG.debug(f"Switched instrument to {index}")
        instrument_model = self._instrument_combo.model()
        self.action.set_instrument(instrument_model._nodes[index])
        self._instrument_index = index
        current_item = self._core.current_item()
        if current_item is not None:
            # we only update the widget if a job is selected from the
            # actions tree
            self.action.apply_instrument()

    @Slot(int)
    def update_action_after_instrument_change(self, index: int):
        if index != self._instrument_index:
            return
        self._needs_updating = True

    @Slot()
    def update_action_on_tab_activation(self):
        if self._needs_updating:
            current_item = self._core.current_item()
            if current_item is not None:
                self.action.apply_instrument()
            self._needs_updating = False

    @classmethod
    def standard_instance(cls):
        action = Action()
        the_tab = cls(
            window,
            name="AvailableJobs",
            model=JobTree(),
            view=ActionsTree(),
            visualiser=Action(),
            layout=partial(
                MultiPanel,
                left_panels=[
                    TextInfo(
                        header="MDANSE Analysis",
                        footer="Look up our Read The Docs page:"
                        + "https://mdanse.readthedocs.io/en/protos/",
                    )
                ],
            ),
            label_text=job_tab_label,
            action=action,
        )
        action._parent_tab = the_tab
        return the_tab

    @classmethod
    def gui_instance(
        cls,
        parent: QWidget,
        name: str,
        session: LocalSession,
        settings,
        logger,
        **kwargs,
    ):
        action = Action(use_preview=True)
        the_tab = cls(
            parent,
            name=name,
            session=session,
            settings=settings,
            logger=logger,
            model=kwargs.get("model", JobTree(filter="Analysis")),
            combo_model=kwargs.get("combo_model", None),
            instrument_model=kwargs.get("instrument_model", None),
            view=ActionsTree(),
            visualiser=action,
            layout=partial(
                MultiPanel,
                left_panels=[
                    TextInfo(
                        header="MDANSE Analysis",
                        footer="Look up our "
                        + '<a href="https://mdanse.readthedocs.io/en/protos/">Read The Docs</a>'
                        + " page.",
                    )
                ],
            ),
            label_text=job_tab_label,
            action=action,
        )
        action.set_settings(the_tab._settings)
        return the_tab


if __name__ == "__main__":
    import sys
    from qtpy.QtWidgets import QApplication, QMainWindow

    app = QApplication(sys.argv)
    window = QMainWindow()
    action = Action()
    the_tab = JobTab(
        window,
        name="AvailableJobs",
        model=JobTree(),
        view=ActionsTree(),
        visualiser=action,
        layout=partial(
            MultiPanel,
            left_panels=[
                TextInfo(
                    header="MDANSE Analysis",
                    footer="Look up our "
                    + '<a href="https://mdanse.readthedocs.io/en/protos/">Read The Docs</a>'
                    + " page.",
                )
            ],
        ),
        label_text=job_tab_label,
        action=action,
    )
    the_tab.update_trajectory_list(["/Users/maciej.bartkowiak/an_example/BLAH.mdt"])
    window.setCentralWidget(the_tab._core)
    window.show()
    app.exec()



#    This file is part of MDANSE_GUI.
#
#    MDANSE_GUI is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program.  If not, see <https://www.gnu.org/licenses/>.
#
from qtpy.QtWidgets import (
    QComboBox,
    QLineEdit,
    QPushButton,
    QVBoxLayout,
    QHBoxLayout,
    QGroupBox,
    QLabel,
    QTextEdit,
    QWidget,
)

from MDANSE.Framework.Configurators.AtomTransmutationConfigurator import AtomTransmuter
from MDANSE.Framework.InputData.HDFTrajectoryInputData import HDFTrajectoryInputData
from MDANSE.Chemistry import ATOMS_DATABASE
from MDANSE_GUI.InputWidgets.AtomSelectionWidget import AtomSelectionWidget
from MDANSE_GUI.InputWidgets.AtomSelectionWidget import SelectionHelper


class TransmutationHelper(SelectionHelper):
    """Generates a string that specifies the atom transmutation.

    Attributes
    ----------
    _helper_title : str
        The title of the helper dialog window.
    """

    _helper_title = "Atom transmutation helper"

    def __init__(
        self,
        transmuter: AtomTransmuter,
        traj_data: tuple[str, HDFTrajectoryInputData],
        field: QLineEdit,
        parent,
        *args,
        **kwargs,
    ):
        """
        Parameters
        ----------
        transmuter : AtomTransmuter
            The MDANSE atom transmuter initialized with the current
            chemical system.
        traj_data : tuple[str, HDFTrajectoryInputData]
            A tuple of the trajectory data used to load the 3D viewer.
        field : QLineEdit
            The QLineEdit field that will need to be updated when
            applying the setting.
        """
        self.transmuter = transmuter
        self.transmutation_textbox = QTextEdit()
        self.transmutation_textbox.setReadOnly(True)
        self.transmutation_combo = QComboBox()
        self.transmutation_combo.addItems(ATOMS_DATABASE.atoms)
        self.transmuter.selector.settings["all"] = False
        super().__init__(
            transmuter.selector,
            traj_data,
            field,
            parent,
            *args,
            **kwargs,
        )
        self.all_selection = False
        self.update_transmutation_textbox()

    def right_widgets(self) -> list[QWidget]:
        """Add the transmutation textbox to the right widgets.

        Returns
        -------
        list[QWidget]
            List of QWidgets to add to the right layout from
            create_layouts.
        """
        widgets = super().right_widgets()
        return widgets + [self.transmutation_textbox]

    def left_widgets(self) -> list[QWidget]:
        """Adds a transmutation widget to the selection helper.

        Returns
        -------
        list[QWidget]
            List of QWidgets to add to the first layout from
            create_layouts.
        """
        widgets = super().left_widgets()
        transmutation = QGroupBox("transmutation")
        transmutation_layout = QVBoxLayout()

        combo_layout = QHBoxLayout()
        label = QLabel("Transmute selection to:")

        combo_layout.addWidget(label)
        combo_layout.addWidget(self.transmutation_combo)
        transmutation_layout.addLayout(combo_layout)

        transmute = QPushButton("Transmute")
        transmutation_layout.addWidget(transmute)
        transmute.clicked.connect(self.apply_transmutation)

        transmutation.setLayout(transmutation_layout)
        return widgets + [transmutation]

    def apply_transmutation(self) -> None:
        """With the selection and the transmutation choice apply the
        transmutation and update the transmutation textbox with the new
        transmutation setting.
        """
        self.transmuter.apply_transmutation(
            self.full_settings, self.transmutation_combo.currentText()
        )
        self.update_transmutation_textbox()

    def update_transmutation_textbox(self) -> None:
        """Update the transmutation textbox with the current transmuter
        setting information.
        """
        map = self.transmuter.get_setting()

        text = [f"Number of atoms transmuted:\n{len(map)}\n\nTransmuted atoms:\n"]
        atoms = self.selector.system.atom_list
        for idx, symbol in map.items():
            text.append(
                f"{idx}  ({atoms[idx].full_name}): {atoms[idx].symbol} -> {symbol}\n"
            )

        self.transmutation_textbox.setText("".join(text))

    def reset(self):
        """Reset the transmuter so that no transmutation are set."""
        super().reset()
        self.transmuter.reset_setting()
        self.update_transmutation_textbox()

    def apply(self) -> None:
        """Set the field of the AtomTransmutationWidget to the currently
        chosen setting in this widget.
        """
        self._field.setText(self.transmuter.get_json_setting())


class AtomTransmutationWidget(AtomSelectionWidget):
    """The atoms transmutation widget."""

    _push_button_text = "Atom transmutation helper"
    _default_value = "{}"
    _tooltip_text = "Specify the atom transmutation that will be used in the analysis. The input is a JSON string, and can be created using the helper dialog."

    def create_helper(
        self, traj_data: tuple[str, HDFTrajectoryInputData]
    ) -> TransmutationHelper:
        """
        Parameters
        ----------
        traj_data : tuple[str, HDFTrajectoryInputData]
            A tuple of the trajectory data used to load the 3D viewer.

        Returns
        -------
        TransmutationHelper
            Create and return the transmutation helper QDialog.
        """
        transmuter = self._configurator.get_transmuter()
        return TransmutationHelper(transmuter, traj_data, self._field, self._base)



#    This file is part of MDANSE_GUI.
#
#    MDANSE_GUI is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program.  If not, see <https://www.gnu.org/licenses/>.
#

from qtpy.QtWidgets import QSpinBox, QLabel
from qtpy.QtCore import Slot

from MDANSE_GUI.InputWidgets.WidgetBase import WidgetBase


suffix_dict = {
    "0": "th order - no interpolation",
    "1": "st order",
    "2": "nd order",
    "3": "rd order",
    "4": "th order",
    "5": "th order",
}


class InterpolationOrderWidget(WidgetBase):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, layout_type="QHBoxLayout", **kwargs)
        source_object = kwargs.get("source_object", None)

        self._field = QSpinBox(self._base)
        self._field.setMaximum(5)
        configurator = self._configurator
        trajectory = configurator._configurable[
            configurator._dependencies["trajectory"]
        ]["instance"]
        if not trajectory.has_variable("velocities"):
            self._field.setMinimum(1)
            self._field.setValue(3)
        label = QLabel("Interpolation order", self._base)
        self.numerator = QLabel("st order")
        self.adjust_numerator(3)

        self._layout.addWidget(label)
        self._layout.addWidget(self._field)
        self._layout.addWidget(self.numerator)
        self.default_labels()
        self.update_labels()
        self.updateValue()
        if self._tooltip:
            tooltip_text = self._tooltip
        else:
            tooltip_text = "The order of the polynomial function used for interpolating velocity values from atom positions. If zero, velocity values present in the trajectory will be used."
        self._field.setToolTip(tooltip_text)
        self.numerator.setToolTip(tooltip_text)
        self._field.valueChanged.connect(self.adjust_numerator)

    def configure_using_default(self):
        """This is too simple to have a default value"""

    def default_labels(self):
        """Each Widget should have a default tooltip and label,
        which will be set in this method, unless specific
        values are provided in the settings of the job that
        is being configured."""
        if self._label_text == "":
            self._label_text = "InterpolationOrderWidget"
        if self._tooltip == "":
            self._tooltip = "The order of the polynomial function used for interpolating velocity values from atom positions. If zero, velocity values present in the trajectory will be used."

    @Slot(int)
    def adjust_numerator(self, order: int):
        text_order = str(order)
        new_numerator = suffix_dict.get(text_order[-1], " - no interpolation")
        self.numerator.setText(new_numerator)

    def get_widget_value(self):
        value = self._field.value()

        return value



#    This file is part of MDANSE_GUI.
#
#    MDANSE_GUI is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program.  If not, see <https://www.gnu.org/licenses/>.
#
from qtpy.QtWidgets import QComboBox

from MDANSE_GUI.InputWidgets.WidgetBase import WidgetBase


class MoleculeWidget(WidgetBase):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        configurator = kwargs.get("configurator", None)
        trajectory_configurator = kwargs.get("trajectory_configurator", None)
        default_option = ""
        if trajectory_configurator is not None:
            option_list = trajectory_configurator[
                "instance"
            ].chemical_system.unique_molecules()
            if len(option_list) > 0:
                default_option = option_list[0]
        else:
            if configurator is None:
                option_list = kwargs.get("choices", [])
            else:
                option_list = configurator.choices
                default_option = configurator.default
        field = QComboBox(self._base)
        field.addItems(option_list)
        field.setCurrentText(default_option)
        field.currentTextChanged.connect(self.updateValue)
        if self._tooltip:
            tooltip_text = self._tooltip
        else:
            tooltip_text = (
                "A single option can be picked out of all the options listed."
            )
        field.setToolTip(tooltip_text)
        self._field = field
        self._layout.addWidget(field)
        self._configurator = configurator
        self.default_labels()
        self.update_labels()
        self.updateValue()

    def configure_using_default(self):
        """This is too complex to have a default value"""

    def default_labels(self):
        """Each Widget should have a default tooltip and label,
        which will be set in this method, unless specific
        values are provided in the settings of the job that
        is being configured."""
        if self._label_text == "":
            self._label_text = "ComboWidget"
        if self._tooltip == "":
            self._tooltip = "You only have one option. Choose wisely."

    def get_widget_value(self):
        return self._field.currentText()



#    This file is part of MDANSE_GUI.
#
#    MDANSE_GUI is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program.  If not, see <https://www.gnu.org/licenses/>.
#
from abc import ABC, abstractmethod


class JobState(ABC):

    _label = "JobState"
    _allowed_actions = []

    def __init__(self, parent):
        self._parent = parent

    @abstractmethod
    def pause(self):
        """Pauses the process"""

    @abstractmethod
    def unpause(self):
        """Resumes the process execution"""

    @abstractmethod
    def start(self):
        """Starts the process"""

    @abstractmethod
    def terminate(self):
        """Instructs the process to stop"""

    @abstractmethod
    def kill(self):
        """Stops the process the hard way"""

    @abstractmethod
    def finish(self):
        """Reach the end of the run successfully"""

    @abstractmethod
    def fail(self):
        """Break down on before finishing"""


class Running(JobState):

    _label = "Running"
    _allowed_actions = [
        "Pause",
        "Terminate",
        "Kill",
    ]

    def pause(self):
        """Pauses the process"""
        self._parent._pause_event.clear()
        self._parent._current_state = self._parent._Paused

    def unpause(self):
        """Resumes the process execution"""

    def start(self):
        """Starts the process"""

    def terminate(self):
        """Instructs the process to stop"""
        self._parent._current_state = self._parent._Aborted

    def kill(self):
        """Stops the process the hard way"""
        self._parent._current_state = self._parent._Aborted

    def finish(self):
        """Reach the end of the run successfully"""
        self._parent.percent_complete = 100
        self._parent._current_state = self._parent._Finished

    def fail(self):
        """Break down on before finishing"""
        self._parent._current_state = self._parent._Failed


class Aborted(JobState):

    _label = "Aborted"
    _allowed_actions = ["Delete"]

    def pause(self):
        """Pauses the process"""

    def unpause(self):
        """Resumes the process execution"""

    def start(self):
        """Starts the process"""

    def terminate(self):
        """Instructs the process to stop"""

    def kill(self):
        """Stops the process the hard way"""

    def finish(self):
        """Reach the end of the run successfully"""

    def fail(self):
        """Break down on before finishing"""


class Failed(JobState):

    _label = "Failed"
    _allowed_actions = ["Delete"]

    def pause(self):
        """Pauses the process"""

    def unpause(self):
        """Resumes the process execution"""

    def start(self):
        """Starts the process"""

    def terminate(self):
        """Instructs the process to stop"""

    def kill(self):
        """Stops the process the hard way"""

    def finish(self):
        """Reach the end of the run successfully"""

    def fail(self):
        """Break down on before finishing"""


class Finished(JobState):

    _label = "Finished"
    _allowed_actions = ["Delete"]

    def pause(self):
        """Pauses the process"""

    def unpause(self):
        """Resumes the process execution"""

    def start(self):
        """Starts the process"""

    def terminate(self):
        """Instructs the process to stop"""

    def kill(self):
        """Stops the process the hard way"""

    def finish(self):
        """Reach the end of the run successfully"""

    def fail(self):
        """Break down on before finishing"""


class Starting(JobState):

    _label = "Starting"
    _allowed_actions = [
        "Pause",
        "Terminate",
        "Kill",
    ]

    def pause(self):
        """Pauses the process"""

    def unpause(self):
        """Resumes the process execution"""

    def start(self):
        """Starts the process"""
        self._parent._current_state = self._parent._Running

    def terminate(self):
        """Instructs the process to stop"""

    def kill(self):
        """Stops the process the hard way"""

    def finish(self):
        """Reach the end of the run successfully"""

    def fail(self):
        """Break down on before finishing"""
        self._parent._current_state = self._parent._Failed


class Paused(JobState):

    _label = "Paused"
    _allowed_actions = [
        "Resume",
        "Terminate",
        "Kill",
    ]

    def pause(self):
        """Pauses the process"""

    def unpause(self):
        """Resumes the process execution"""
        self._parent._pause_event.set()
        self._parent._current_state = self._parent._Running

    def start(self):
        """Starts the process"""

    def terminate(self):
        """Instructs the process to stop"""
        self._parent._current_state = self._parent._Aborted

    def kill(self):
        """Stops the process the hard way"""
        self._parent._current_state = self._parent._Aborted

    def finish(self):
        """Reach the end of the run successfully"""
        self._parent.percent_complete = 100
        self._parent._current_state = self._parent._Finished

    def fail(self):
        """Break down on before finishing"""



#    This file is part of MDANSE_GUI.
#
#    MDANSE_GUI is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program.  If not, see <https://www.gnu.org/licenses/>.
#

import multiprocessing

from qtpy.QtWidgets import QComboBox, QSpinBox
from qtpy.QtCore import Slot

from MDANSE_GUI.InputWidgets.WidgetBase import WidgetBase


class RunningModeWidget(WidgetBase):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._totalNumberOfProcessors = multiprocessing.cpu_count()
        self._last_numproc = None
        default_value = self._configurator.default
        self._value = default_value
        self.mode_box = QComboBox(self._base)
        self.mode_box.addItems(self._configurator.availablesModes)
        self.mode_box.setCurrentText(self._configurator.availablesModes[0])
        self._field = QSpinBox(self._base)
        self._field.setValue(1)
        self._field.setMinimum(1)
        self._field.setMaximum(multiprocessing.cpu_count())
        self._field.setEnabled(False)
        self._layout.addWidget(self.mode_box)
        self._layout.addWidget(self._field)
        self.mode_box.currentIndexChanged.connect(self.mode_changed)
        self._field.valueChanged.connect(self.numproc_changed)
        self.update_labels()
        self.updateValue()
        if self._tooltip:
            tooltip_text = self._tooltip
        else:
            tooltip_text = "Choose if the job should run on one or more cores."
        for wid in [self._field, self.mode_box]:
            wid.setToolTip(tooltip_text)

    def configure_using_default(self):
        """This is too complex to have a default value"""

    @Slot()
    def mode_changed(self):
        mode = self.mode_box.currentText()
        if self._last_numproc is None:
            nextval = self._totalNumberOfProcessors
        else:
            nextval = self._last_numproc
        if mode == "single-core":
            self._field.setValue(1)
            self._field.setEnabled(False)
        else:
            self._field.setEnabled(True)
            self._field.setValue(nextval)

    @Slot()
    def numproc_changed(self):
        mode = self.mode_box.currentText()
        numproc = self._field.value()
        if mode == "single-core":
            return
        self._last_numproc = numproc

    def get_widget_value(self):
        mode = self.mode_box.currentText()
        numproc = self._field.value()
        if mode == "single-core":
            value = (mode,)
        else:
            value = (mode, numproc)
        return value



import pytest
import tempfile

from MDANSE_GUI.Tabs.Models.GeneralModel import GeneralModel


@pytest.fixture(scope="module")
def model_instance():
    instance = GeneralModel()
    instance.append_object((5, "number"))
    instance.append_object(([], "something"))
    instance.append_object(("However", "text"))
    return instance


def test_item_access(model_instance: GeneralModel):
    assert model_instance._nodes[0] == 5
    assert model_instance._nodes[1] == []
    assert model_instance._nodes[2] == "However"


def test_labels(model_instance: GeneralModel):
    for num, textval in enumerate(["number", "something", "text"]):
        index = model_instance.index(num, 0)
        std_item = model_instance.itemFromIndex(index)
        label = std_item.text()
        assert label == textval


def test_removal(model_instance: GeneralModel):
    model_instance.removeRow(1)
    assert model_instance._nodes[0] == 5
    assert model_instance._nodes[2] == "However"
    assert len(model_instance._node_numbers) == 2
    assert len(model_instance._nodes) == 2



#    This file is part of MDANSE_GUI.
#
#    MDANSE_GUI is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program.  If not, see <https://www.gnu.org/licenses/>.
#
from typing import Union

from qtpy.QtWidgets import QTreeView, QAbstractItemView, QApplication
from qtpy.QtCore import Signal, Slot, QModelIndex, Qt, QMimeData
from qtpy.QtGui import QMouseEvent, QDrag

from MDANSE.MLogging import LOG

from MDANSE_GUI.DataViewModel.TrajectoryHolder import DataTreeItem
from MDANSE_GUI.DataViewModel.ActionsHolder import ActionsHolder
from MDANSE_GUI.Tabs.Visualisers.Action import Action
from MDANSE_GUI.Tabs.Visualisers.TextInfo import TextInfo


class ActionsTree(QTreeView):
    jobname_selected = Signal(str)
    execute_action = Signal(object)
    item_details = Signal(object)
    error = Signal(str)

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.setHeaderHidden(True)
        self.setEditTriggers(QAbstractItemView.EditTrigger.NoEditTriggers)
        self.click_position = None

        self.clicked.connect(self.on_select_action)
        self.doubleClicked.connect(self.pop_action_dialog)
        self.clicked.connect(self.item_picked)

    def mousePressEvent(self, e: QMouseEvent) -> None:
        self.click_position = e.position()
        if self.model() is None:
            return None
        return super().mousePressEvent(e)

    def on_select_action(self, index):
        model = self.model()
        item = model.itemFromIndex(index)
        text = item.text()
        LOG.info(f"tree: clicked on {text}")
        self.jobname_selected.emit(text)

    def pop_action_dialog(self, index):
        model = self.model()
        item = model.itemFromIndex(index)
        # debug
        text = item.text()
        LOG.info(f"About to execute action {text}")
        #
        number = item.data(Qt.ItemDataRole.UserRole)
        LOG.info(f"Node number is {number}")
        if number is None:
            return
        action = model._values[number]
        self.execute_action.emit(action)

    @Slot(DataTreeItem)
    def showValidActions(self, item: DataTreeItem):
        LOG.info(f"Creating model from {item}")
        new_model = ActionsHolder(item)
        self.setModel(new_model)

    @Slot(QModelIndex)
    def item_picked(self, index: QModelIndex):
        model = self.model()
        node_number = model.itemFromIndex(index).data(Qt.ItemDataRole.UserRole)
        try:
            job_description = model._docstrings[node_number]
        except KeyError:
            job_description = "No further information"
        self.item_details.emit(job_description)  # this should emit the job name

    def connect_to_visualiser(self, visualiser: Union[Action, TextInfo]) -> None:
        """Connect to a visualiser.

        Parameters
        ----------
        visualiser : Action or TextInfo
            A visualiser to connect to this view.
        """
        if isinstance(visualiser, Action):
            self.jobname_selected.connect(visualiser.update_panel)
        elif isinstance(visualiser, TextInfo):
            self.item_details.connect(visualiser.update_panel)
        else:
            raise NotImplementedError(
                f"Unable to connect view {type(self)} to visualiser "
                f"{type(visualiser)}"
            )



import fnmatch
import os
import sys

import numpy

from setuptools import setup, find_packages
from distutils.sysconfig import get_config_vars
from distutils.util import convert_path

try:
    import sphinx
except ImportError:
    sphinx = None


#################################
# Modules variables
#################################
EXCLUDE = ["*.py", "*.pyc", "*$py.class", "*~", ".*", "*.bak", "*.so", "*.pyd"]

EXCLUDE_DIRECTORIES = (
    ".*",
    "CVS",
    "_darcs",
    "./build",
    "*svn",
    "./dist",
    "EGG-INFO",
    "*.egg-info",
)

EXTENSIONS_PATH = "Extensions"

INCLUDE_DIR = [numpy.get_include()]

QHULL_DIR = os.path.join("Extensions", "qhull_lib")

QHULL_INCLUDE_DIR = (
    INCLUDE_DIR
    + [EXTENSIONS_PATH]
    + [os.path.join(QHULL_DIR, "ext")]
    + [os.path.join(QHULL_DIR, "src")]
)

#################################
# Helper function
#################################


def is_package(path):
    return os.path.isdir(path) and os.path.isfile(os.path.join(path, "__init__.py"))


def find_package_data(
    where=".",
    package="",
    exclude=EXCLUDE,
    exclude_directories=EXCLUDE_DIRECTORIES,
    only_in_packages=True,
    show_ignored=False,
):

    out = {}
    stack = [(convert_path(where), "", package, only_in_packages)]
    while stack:
        where, prefix, package, only_in_packages = stack.pop(0)
        for name in os.listdir(where):
            fn = os.path.join(where, name)
            if os.path.isdir(fn):
                bad_name = False
                for pattern in exclude_directories:
                    if (
                        fnmatch.fnmatchcase(name, pattern)
                        or fn.lower() == pattern.lower()
                    ):
                        bad_name = True
                        if show_ignored:
                            print >> sys.stderr, (
                                "Directory %s ignored by pattern %s" % (fn, pattern)
                            )
                        break
                if bad_name:
                    continue
                if os.path.isfile(os.path.join(fn, "__init__.py")) and not prefix:
                    if not package:
                        new_package = name
                    else:
                        new_package = package + "." + name
                    stack.append((fn, "", new_package, False))
                else:
                    stack.append((fn, prefix + name + "/", package, only_in_packages))
            elif package or not only_in_packages:
                # is a file
                bad_name = False
                for pattern in exclude:
                    if (
                        fnmatch.fnmatchcase(name, pattern)
                        or fn.lower() == pattern.lower()
                    ):
                        bad_name = True
                        if show_ignored:
                            print >> sys.stderr, (
                                "File %s ignored by pattern %s" % (fn, pattern)
                            )
                        break
                if bad_name:
                    continue
                out.setdefault(package, []).append(prefix + name)

    return out


def find_data(
    where=".", exclude=EXCLUDE, exclude_directories=EXCLUDE_DIRECTORIES, prefix=""
):

    out = {}
    stack = [convert_path(where)]
    while stack:
        where = stack.pop(0)
        for name in os.listdir(where):
            fn = os.path.join(where, name)
            d = os.path.join(prefix, os.path.dirname(fn))
            if os.path.isdir(fn):
                stack.append(fn)
            else:
                bad_name = False
                for pattern in exclude:
                    if (
                        fnmatch.fnmatchcase(name, pattern)
                        or fn.lower() == pattern.lower()
                    ):
                        bad_name = True
                        break
                if bad_name:
                    continue
                out.setdefault(d, []).append(fn)

    out = [(k, v) for k, v in out.items()]

    return out


#################################
# Documentation
#################################

if sphinx:

    try:
        from sphinx.ext.apidoc import main as sphinx_apidoc_main
    except ImportError:
        from sphinx.apidoc import main as sphinx_apidoc_main

    #     class mdanse_build_doc(sphinx.setup_command.BuildDoc):
    class mdanse_build_doc:

        def run(self):

            build = self.get_finalized_command("build")

            buildDir = os.path.abspath(build.build_lib)

            if not os.path.exists(buildDir):
                raise IOError("build command must be performed prior building the doc")

            sys.path.insert(0, buildDir)

            sphinxDir = os.path.abspath(
                os.path.join(build.build_base, "sphinx", self.doctype)
            )

            if not os.path.exists(sphinxDir):
                os.makedirs(sphinxDir)

            metadata = self.distribution.metadata
            args = [
                "-F",
                "--separate",
                "-H%s" % metadata.name,
                "-A%s" % metadata.author,
                "-R%s" % metadata.version,
                "-o%s" % sphinxDir,
                os.path.join(buildDir, "MDANSE"),
                os.path.join(buildDir, "MDANSE", "Externals"),
            ]

            # /!\ apidoc.main is deprecated. The API has been broken in sphinx 1.7.0, see https://github.com/sphinx-doc/sphinx/issues/4615
            if int(sphinx.__version__.split(".")[1]) <= 6:
                args.insert(0, "")

            sphinx_apidoc_main(args)

            currentDirectory = os.getcwd()

            import shutil

            shutil.copy(
                os.path.join(currentDirectory, "Doc", "conf_%s.py" % self.doctype),
                os.path.join(sphinxDir, "conf.py"),
            )
            shutil.copy(
                os.path.join(currentDirectory, "Doc", "mdanse_logo.png"),
                os.path.join(sphinxDir, "_static"),
            )
            shutil.copy(
                os.path.join(currentDirectory, "Doc", "layout.html"),
                os.path.join(sphinxDir, "_templates"),
            )

            # The directory where the rst files are located.
            self.source_dir = sphinxDir
            # The directory where the conf.py file is located.
            self.config_dir = self.source_dir

            # The directory where the documentation will be built
            self.build_dir = os.path.join(buildDir, "MDANSE", "Doc", self.doctype)

            if isinstance(self.builder, str):
                builders = [self.builder]
            else:
                builders = self.builder

            for builder in builders:
                self.builder_target_dir = os.path.join(self.build_dir, builder)
                sphinx.setup_command.BuildDoc.finalize_options(self)
                sphinx.setup_command.BuildDoc.run(self)

            sys.path.pop(0)

    class mdanse_build_help(mdanse_build_doc):

        doctype = "help"

    class mdanse_build_api(mdanse_build_doc):

        doctype = "api"


#################################
# Extensions section
#################################

if "linux" in sys.platform:
    (opt,) = get_config_vars("OPT")
    os.environ["OPT"] = " ".join(
        flag for flag in opt.split() if flag != "-Wstrict-prototypes"
    )

CMDCLASS = {}

if sphinx:
    CMDCLASS["build_api"] = mdanse_build_api
    CMDCLASS["build_help"] = mdanse_build_help

#################################
# The setup section
#################################

setup(
    name="MDANSE GUI",
    packages=find_packages("Src"),
    package_dir={"": "Src"},
    platforms=["Unix", "Windows"],
    cmdclass=CMDCLASS,
)



#    This file is part of MDANSE_GUI.
#
#    MDANSE_GUI is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program.  If not, see <https://www.gnu.org/licenses/>.
#
from qtpy.QtWidgets import QLineEdit, QPushButton, QFileDialog, QComboBox
from qtpy.QtCore import Slot
from ase.io.formats import filetype

from MDANSE.MLogging import LOG
from MDANSE.Framework.Configurators.AseInputFileConfigurator import (
    AseInputFileConfigurator,
)

from MDANSE_GUI.InputWidgets.WidgetBase import WidgetBase


class AseInputFileWidget(WidgetBase):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        configurator = kwargs.get("configurator", None)
        if configurator is not None:
            default_value = configurator.default
        else:
            default_value = ""
        try:
            parent = kwargs.get("parent", None)
            self.default_path = parent.default_path
        except KeyError:
            self.default_path = "."
            LOG.error("KeyError in InputFileWidget - can't get default path.")
        except AttributeError:
            self.default_path = "."
            LOG.error("AttributeError in InputFileWidget - can't get default path.")
        default_value = kwargs.get("default", "")
        if self._tooltip:
            tooltip_text = self._tooltip
        else:
            tooltip_text = "Specify a path to an existing file."
        self._qt_file_association = kwargs.get("wildcard", "")
        combo = QComboBox(self._base)
        combo.addItems(AseInputFileConfigurator._allowed_formats)
        combo.setEditable(False)
        self._type_combo = combo
        field = QLineEdit(self._base)
        self._field = field
        field.textChanged.connect(self.updateValue)
        field.setText(str(default_value))
        field.setToolTip(tooltip_text)
        self._layout.addWidget(field)
        self._layout.addWidget(combo)
        button = QPushButton("Browse", self._base)
        button.clicked.connect(self.valueFromDialog)
        self._layout.addWidget(button)
        self._configurator = configurator
        self._file_dialog = QFileDialog.getOpenFileName
        self.updateValue()

    @Slot()
    def valueFromDialog(self):
        """A Slot defined to allow the GUI to be updated based on
        the new path received from a FileDialog.
        This will start a FileDialog, take the resulting path,
        and emit a signal to update the value show by the GUI.
        """
        new_value = self._file_dialog(
            self.parent(),  # the parent of the dialog
            "Load file",  # the label of the window
            self.default_path,  # the initial search path
            self._qt_file_association,  # text string specifying the file name filter.
        )
        if new_value is not None:
            self._field.setText(new_value[0])
            self.updateValue()
            try:
                type_guess = filetype(new_value[0])
            except:
                type_guess = "guess"
            self._type_combo.setCurrentText(type_guess)

    def get_widget_value(self):
        """Collect the results from the input widgets and return the value."""
        return self._field.text(), self._type_combo.currentText()



#    This file is part of MDANSE_GUI.
#
#    MDANSE_GUI is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program.  If not, see <https://www.gnu.org/licenses/>.
#
from typing import List

import numpy as np
from scipy.spatial import cKDTree as KDTree

from qtpy import QtWidgets
from qtpy.QtCore import Signal, Slot
from qtpy.QtWidgets import QSizePolicy

import vtk
from vtkmodules.qt.QVTKRenderWindowInteractor import QVTKRenderWindowInteractor
from vtkmodules.vtkRenderingAnnotation import vtkAxesActor

from MDANSE.Framework.InputData.HDFTrajectoryInputData import HDFTrajectoryInputData
from MDANSE.Chemistry import ATOMS_DATABASE as CHEMICAL_ELEMENTS
from MDANSE.MLogging import LOG

from MDANSE_GUI.MolecularViewer.readers import hdf5wrapper
from MDANSE_GUI.MolecularViewer.Dummy import PyConnectivity
from MDANSE_GUI.MolecularViewer.Contents import TrajectoryAtomData
from MDANSE_GUI.MolecularViewer.AtomProperties import (
    AtomProperties,
    ndarray_to_vtkarray,
)


def array_to_3d_imagedata(data, spacing):
    nx = data.shape[0]
    ny = data.shape[1]
    nz = data.shape[2]
    image = vtk.vtkImageData()
    image.SetDimensions(nx, ny, nz)
    dx, dy, dz = spacing
    image.SetSpacing(dx, dy, dz)
    image.SetExtent(0, nx - 1, 0, ny - 1, 0, nz - 1)
    if vtk.vtkVersion.GetVTKMajorVersion() < 6:
        image.SetScalarTypeToDouble()
        image.SetNumberOfScalarComponents(1)
    else:
        image.AllocateScalars(vtk.VTK_DOUBLE, 1)

    for i in range(nx):
        for j in range(ny):
            for k in range(nz):
                image.SetScalarComponentFromDouble(i, j, k, 0, data[i, j, k])

    return image


class MolecularViewer(QtWidgets.QWidget):
    """This class implements a molecular viewer."""

    new_max_frames = Signal(int)

    def __init__(self):
        super(MolecularViewer, self).__init__()

        self._scale_factor = 0.8

        self._datamodel = None

        self._iren = QVTKRenderWindowInteractor(self)

        def dummy_method(self, ev=None):
            pass

        setattr(self._iren, "keyPressEvent", dummy_method)

        self._renderer = vtk.vtkRenderer()

        self._iren.GetRenderWindow().AddRenderer(self._renderer)

        self._iren.GetRenderWindow().SetPosition((0, 0))

        self._iren.GetInteractorStyle().SetCurrentStyleToTrackballCamera()

        self._iren.Enable()

        self._iren.GetRenderWindow()

        self.axes_actor = vtkAxesActor()
        self.axes_actor.AxisLabelsOn()
        self.axes_actor.SetShaftTypeToCylinder()
        self.axes_widget = vtk.vtkOrientationMarkerWidget()
        self.axes_widget.SetOrientationMarker(self.axes_actor)
        self.axes_widget.SetInteractor(self._iren.GetRenderWindow().GetInteractor())
        self.axes_widget.SetViewport(0.0, 0.0, 0.25, 0.25)
        self.axes_widget.SetEnabled(True)
        self.axes_widget.InteractiveOff()

        layout = QtWidgets.QStackedLayout(self)
        layout.addWidget(self._iren)
        self.setSizePolicy(QSizePolicy.Policy.Minimum, QSizePolicy.Policy.Minimum)

        # create camera
        self._camera = vtk.vtkCamera()
        # associate camera to renderer
        self._renderer.SetActiveCamera(self._camera)
        self._camera.SetFocalPoint(0, 0, 0)
        self._camera.SetPosition(0, 0, 20)

        self._n_atoms = 0
        self._n_frames = 0
        self._resolution = 0

        self._atoms_visible = True
        self._bonds_visible = True
        self._axes_visible = True
        self._cell_visible = True

        self._iren.Initialize()

        self._atoms = []

        self._polydata = None
        self._polydata_bonds_exist = False
        self._uc_polydata = None

        self._surface = None

        self._reader = None

        self._current_frame = 0

        self._colour_manager = AtomProperties()
        self.dummy_size = 0.0

        self.reset_camera = False

    def setDataModel(self, datamodel: TrajectoryAtomData):
        self._datamodel = datamodel

    def _new_trajectory_object(self, fname: str, data: HDFTrajectoryInputData):
        reader = hdf5wrapper.HDF5Wrapper(fname, data.trajectory, data.chemical_system)
        self.set_reader(reader)

    @Slot(str)
    def _new_trajectory(self, fname: str):
        data = HDFTrajectoryInputData(fname)
        reader = hdf5wrapper.HDF5Wrapper(fname, data.trajectory, data.chemical_system)
        self.set_reader(reader)

    @Slot(float)
    def _new_scaling(self, scale_factor: float):
        self._scale_factor = scale_factor
        self.update_renderer()

    def _new_visibility(self, flags: List[bool]):
        self._atoms_visible = flags[0]
        self._bonds_visible = flags[1]
        self._axes_visible = flags[2]
        self._cell_visible = flags[3]
        self.axes_widget.SetEnabled(self._axes_visible)
        result = self.set_coordinates(self._current_frame)
        if result is False:
            self.update_renderer()

    @Slot()
    def _new_atom_parameters(self):
        if self._polydata is None or self._datamodel is None:
            return

        # we need to add the new colours to LUT
        # then assign new colour NUMBERS to _atom_colours

        # we also need to assign new atom radii to _atom_scales

        scalars = ndarray_to_vtkarray(
            self._atom_colours, self._atom_scales, self._n_atoms
        )

        self._polydata.GetPointData().SetScalars(scalars)

        self.update_renderer()

    def _draw_isosurface(self, index):
        """Draw the isosurface of an atom with the given index"""

        return None  # we are not ready for this

        if self._surface is not None:
            self.on_clear_atomic_trace()

        LOG.info("Computing isosurface ...")

        initial_coords = self._reader.read_frame(0)
        coords, lower_bounds, upper_bounds = self._reader.read_atom_trajectory(index)
        spacing, self._atomic_trace_histogram = histogram_3d(
            coords, lower_bounds, upper_bounds, 100, 100, 100
        )

        self._image = array_to_3d_imagedata(self._atomic_trace_histogram, spacing)
        isovalue = self._atomic_trace_histogram.mean()

        self._isocontour = vtk.vtkMarchingContourFilter()
        self._isocontour.UseScalarTreeOn()
        self._isocontour.ComputeNormalsOn()
        if vtk.vtkVersion.GetVTKMajorVersion() < 6:
            self._isocontour.SetInput(self.image)
        else:
            self._isocontour.SetInputData(self._image)
        self._isocontour.SetValue(0, isovalue)

        self._depthSort = vtk.vtkDepthSortPolyData()
        self._depthSort.SetInputConnection(self._isocontour.GetOutputPort())
        self._depthSort.SetDirectionToBackToFront()
        self._depthSort.SetVector(1, 1, 1)
        self._depthSort.SetCamera(self._camera)
        self._depthSort.SortScalarsOn()
        self._depthSort.Update()

        mapper = vtk.vtkPolyDataMapper()
        mapper.SetInputConnection(self._depthSort.GetOutputPort())
        mapper.ScalarVisibilityOff()
        mapper.Update()

        self._surface = vtk.vtkActor()
        self._surface.SetMapper(mapper)
        self._surface.GetProperty().SetColor((0, 0.5, 0.75))
        self._surface.GetProperty().SetOpacity(0.5)
        self._surface.PickableOff()

        self._surface.GetProperty().SetRepresentationToSurface()

        self._surface.GetProperty().SetInterpolationToGouraud()
        self._surface.GetProperty().SetSpecular(0.4)
        self._surface.GetProperty().SetSpecularPower(10)

        self._renderer.AddActor(self._surface)

        self._surface.SetPosition(
            lower_bounds[0, 0], lower_bounds[0, 1], lower_bounds[0, 2]
        )

        self._iren.Render()

        LOG.info("... done")

    def create_all_actors(self):
        actors = []
        if self._polydata is None:
            return actors

        line_actor, ball_actor = self.create_traj_actors(self._polydata)
        if self._cell_visible:
            uc_actor = self.create_uc_actor()
            actors.append(uc_actor)
        if self._bonds_visible and self._polydata_bonds_exist:
            actors.append(line_actor)
        if self._atoms_visible:
            actors.append(ball_actor)
        return actors

    def create_uc_actor(self):
        uc_mapper = vtk.vtkPolyDataMapper()
        if vtk.vtkVersion.GetVTKMajorVersion() < 6:
            uc_mapper.SetInput(self._uc_polydata)
        else:
            uc_mapper.SetInputData(self._uc_polydata)
        uc_mapper.ScalarVisibilityOn()
        uc_actor = vtk.vtkLODActor()
        uc_actor.GetProperty().SetLineWidth(3 * self._scale_factor)
        uc_actor.SetMapper(uc_mapper)
        return uc_actor

    def create_traj_actors(self, polydata, line_opacity=1.0, ball_opacity=1.0):
        line_mapper = vtk.vtkPolyDataMapper()
        if vtk.vtkVersion.GetVTKMajorVersion() < 6:
            line_mapper.SetInput(polydata)
        else:
            line_mapper.SetInputData(polydata)

        line_mapper.SetLookupTable(self._colour_manager._lut)
        line_mapper.ScalarVisibilityOn()
        line_mapper.ColorByArrayComponent("scalars", 1)
        line_actor = vtk.vtkLODActor()
        line_actor.GetProperty().SetLineWidth(3 * self._scale_factor)
        line_actor.SetMapper(line_mapper)
        line_actor.GetProperty().SetAmbient(0.2)
        line_actor.GetProperty().SetDiffuse(0.5)
        line_actor.GetProperty().SetSpecular(0.3)
        line_actor.GetProperty().SetOpacity(line_opacity)

        temp_radius = float(1.0 * self._scale_factor)
        sphere = vtk.vtkSphereSource()
        sphere.SetCenter(0, 0, 0)
        sphere.SetRadius(temp_radius)
        sphere.SetThetaResolution(self._resolution)
        sphere.SetPhiResolution(self._resolution)
        glyph = vtk.vtkGlyph3D()
        if vtk.vtkVersion.GetVTKMajorVersion() < 6:
            glyph.SetInput(polydata)
        else:
            glyph.SetInputData(polydata)

        temp_scale = float(0.2 * self._scale_factor)
        glyph.SetScaleModeToScaleByScalar()
        glyph.SetColorModeToColorByScalar()
        glyph.SetScaleFactor(temp_scale)
        glyph.SetSourceConnection(sphere.GetOutputPort())
        glyph.SetIndexModeToScalar()
        sphere_mapper = vtk.vtkPolyDataMapper()
        sphere_mapper.SetLookupTable(self._colour_manager._lut)
        sphere_mapper.SetScalarRange(polydata.GetScalarRange())
        sphere_mapper.SetInputConnection(glyph.GetOutputPort())
        sphere_mapper.ScalarVisibilityOn()
        sphere_mapper.ColorByArrayComponent("scalars", 1)
        ball_actor = vtk.vtkLODActor()
        ball_actor.SetMapper(sphere_mapper)
        ball_actor.GetProperty().SetAmbient(0.2)
        ball_actor.GetProperty().SetDiffuse(0.5)
        ball_actor.GetProperty().SetSpecular(0.3)
        ball_actor.GetProperty().SetOpacity(ball_opacity)
        ball_actor.SetNumberOfCloudPoints(30000)
        return [line_actor, ball_actor]

    def clear_trajectory(self):
        """Clear the vtk scene from atoms and bonds actors."""

        if not hasattr(self, "_actors"):
            return
        if self._actors is None:
            return

        self._actors.VisibilityOff()
        self._actors.ReleaseGraphicsResources(self.get_render_window())
        self._renderer.RemoveActor(self._actors)

        del self._actors

    def clear_panel(self) -> None:
        """Clears the Molecular Viewer panel"""
        self.clear_trajectory()

        self._reader = None

        # set everything to some empty/zero value
        self._n_atoms = 0
        self._n_frames = 0
        self.new_max_frames.emit(0)
        self._atoms = []
        self._atom_colours = []
        self._current_frame = 0
        self.reset_all_polydata()

        self.update_renderer()

        # clear the atom properties table
        self._colour_manager.removeRows(0, self._colour_manager.rowCount())

    def reset_all_polydata(self):
        self._polydata = vtk.vtkPolyData()
        self._uc_polydata = vtk.vtkPolyData()

    def update_all_polydata(self):
        self.update_polydata()
        self.update_uc_polydata()

    def update_polydata(self):
        coords = self._reader.read_frame(self._current_frame)

        if self._atoms_visible or self._bonds_visible:
            atoms = vtk.vtkPoints()
            atoms.SetNumberOfPoints(self._n_atoms)
            for i in range(self._n_atoms):
                x, y, z = coords[i, :]
                atoms.SetPoint(i, x, y, z)

            self._polydata.SetPoints(atoms)

        if self._bonds_visible:
            # do not bond atoms to dummy atoms
            not_du = np.array(
                [
                    i
                    for i, at in enumerate(self._reader.atom_types)
                    if CHEMICAL_ELEMENTS.get_atom_property(at, "element") != "dummy"
                ]
            )
            rs = coords[not_du]
            covs = np.array(
                [
                    CHEMICAL_ELEMENTS.get_atom_property(at, "covalent_radius")
                    for at in self._reader.atom_types
                ]
            )[not_du]

            bonds, bonds_exist = self.create_bond_cell_array(rs, covs, not_du)
            if bonds_exist:
                self._polydata.SetLines(bonds)
                self._polydata_bonds_exist = True
                return

        self._polydata_bonds_exist = False

    def create_bond_cell_array(self, rs, covs, not_du, tolerance=0.04):
        # determine and set bonds without PBC applied
        tree = KDTree(rs)
        bonds = vtk.vtkCellArray()
        contacts = tree.query_ball_tree(tree, 2 * np.max(covs) + tolerance)
        n_bonds = 0
        for i, idxs in enumerate(contacts):
            if len(idxs) == 0:
                continue
            diff = rs[i] - rs[idxs]
            dist = np.sum(diff * diff, axis=1)
            sum_radii = (covs[i] + covs[idxs] + tolerance) ** 2
            js = np.array(idxs)[(0 < dist) & (dist < sum_radii)]
            js = not_du[js[i < js]]
            n_bonds += len(js)
            for j in js:
                line = vtk.vtkLine()
                line.GetPointIds().SetId(0, not_du[i])
                line.GetPointIds().SetId(1, j)
                bonds.InsertNextCell(line)
        return bonds, n_bonds > 0

    def update_uc_polydata(self):
        uc = self._reader.read_pbc(self._current_frame)
        if self._cell_visible and uc is not None:
            # update the unit cell
            a = uc.a_vector
            b = uc.b_vector
            c = uc.c_vector
            uc_points = vtk.vtkPoints()
            uc_points.SetNumberOfPoints(8)
            for i, v in enumerate([[0, 0, 0], a, b, c, a + b, a + c, b + c, a + b + c]):
                x, y, z = v
                uc_points.SetPoint(i, x, y, z)
            self._uc_polydata.SetPoints(uc_points)

            uc_lines = vtk.vtkCellArray()
            for i, j in [
                (0, 1),
                (0, 2),
                (0, 3),
                (1, 4),
                (1, 5),
                (4, 7),
                (2, 4),
                (2, 6),
                (5, 7),
                (3, 5),
                (3, 6),
                (6, 7),
            ]:
                line = vtk.vtkLine()
                line.GetPointIds().SetId(0, i)
                line.GetPointIds().SetId(1, j)
                uc_lines.InsertNextCell(line)
            self._uc_polydata.SetLines(uc_lines)

    def get_atom_index(self, pid):
        """Return the atom index from the vtk data point index.

        Args:
            pid (int): the data point index
        """

        _, _, idx = (
            self.glyph.GetOutput().GetPointData().GetArray("scalars").GetTuple3(pid)
        )

        return int(idx)

    def get_render_window(self):
        """Returns the render window."""
        return self._iren.GetRenderWindow()

    @property
    def iren(self):
        return self._iren

    def on_change_atomic_trace_opacity(self, opacity):
        """Event handler called when the opacity level is changed."""

        if self._surface is None:
            return

        self._surface.GetProperty().SetOpacity(opacity)
        self._iren.Render()

    def on_change_atomic_trace_isocontour_level(self, level):
        """Event handler called when the user change the isocontour level."""

        if self._surface is None:
            return

        self._isocontour.SetValue(0, level)
        self._isocontour.Update()
        self._iren.Render()

    def on_change_atomic_trace_rendering_type(self, rendering_type):
        """Event handler called when the user change the rendering type for the atomic trace."""

        if self._surface is None:
            return

        if rendering_type == "wireframe":
            self._surface.GetProperty().SetRepresentationToWireframe()
        elif rendering_type == "surface":
            self._surface.GetProperty().SetRepresentationToSurface()
        elif rendering_type == "points":
            self._surface.GetProperty().SetRepresentationToPoints()
            self._surface.GetProperty().SetPointSize(3)
        else:
            return

        self._iren.Render()

    def on_clear_atomic_trace(self):
        """Event handler called when the user select the 'Atomic trace -> Clear' main menu item"""

        if self._surface is None:
            return

        self._surface.VisibilityOff()
        self._surface.ReleaseGraphicsResources(self._iren.GetRenderWindow())
        self._renderer.RemoveActor(self._surface)
        self._iren.Render()

        self._surface = None

    def on_open_atomic_trace_settings_dialog(self):
        """Event handler which pops the atomic trace settings."""

        if self._surface is None:
            return

        hist_min = self._atomic_trace_histogram.min()
        hist_max = self._atomic_trace_histogram.max()
        hist_mean = self._atomic_trace_histogram.mean()

        dlg = AtomicTraceSettingsDialog(hist_min, hist_max, hist_mean, self)

        dlg.rendering_type_changed.connect(self.on_change_atomic_trace_rendering_type)
        dlg.opacity_changed.connect(self.on_change_atomic_trace_opacity)
        dlg.isocontour_level_changed.connect(
            self.on_change_atomic_trace_isocontour_level
        )

        dlg.show()

    def on_show_atomic_trace(self):
        if self._previously_picked_atom is None:
            LOG.warning("No atom selected for computing atomic trace")
            return

        self._draw_isosurface(self._previously_picked_atom[0])

    @property
    def renderer(self):
        return self._renderer

    def set_connectivity_builder(self, coords, covalent_radii):
        # Compute the bounding box of the system
        lower_bound = coords.min(axis=0)
        upper_bound = coords.max(axis=0)

        # Enlarge it a bit to not miss any atom
        lower_bound -= 1.0e-6
        upper_bound += 1.0e-6

        # Initializes the octree used to build the connectivity
        self._connectivity_builder = PyConnectivity(lower_bound, upper_bound, 0, 10, 18)

        # Add the points to the octree
        for index, xyz, radius in zip(range(self._n_atoms), coords, covalent_radii):
            self._connectivity_builder.add_point(index, xyz, radius)

    @Slot(int)
    def set_coordinates(self, frame: int):
        """Sets a new configuration.

        @param frame: the configuration number
        @type frame: integer
        """
        if self._reader is None:
            return False

        self._current_frame = frame % self._reader.n_frames

        # update the atoms
        self.update_all_polydata()

        # Update the view.
        self.update_renderer()

    def set_reader(self, reader, frame=0):
        """Set the trajectory at a given frame

        Args:
            reader (IReader): the trajectory object
            frame (int): the selected frame
        """

        if (self._reader is not None) and (reader.filename == self._reader.filename):
            return

        self.reset_camera = True
        self.clear_trajectory()

        self._reader = reader

        self._n_atoms = self._reader.n_atoms
        self._n_frames = self._reader.n_frames

        self.new_max_frames.emit(self._n_frames - 1)

        self._atoms = self._reader.atom_types

        # Hack for reducing objects resolution when the system is big
        self._resolution = int(np.sqrt(3000000.0 / self._n_atoms))
        self._resolution = 10 if self._resolution > 10 else self._resolution
        self._resolution = 4 if self._resolution < 4 else self._resolution

        self._atom_colours = self._colour_manager.reinitialise_from_database(
            self._atoms, CHEMICAL_ELEMENTS, self.dummy_size
        )
        # this returs a list of indices, mapping colours to atoms

        self._atom_scales = np.array(
            [
                CHEMICAL_ELEMENTS.get_atom_property(at, "vdw_radius")
                for at in self._atoms
            ]
        ).astype(np.float32)

        scalars = ndarray_to_vtkarray(
            self._atom_colours, self._atom_scales, self._n_atoms
        )

        self.reset_all_polydata()
        self._polydata.GetPointData().SetScalars(scalars)

        self._current_frame = frame
        self._colour_manager.onNewValues()

    @Slot(object)
    def take_atom_properties(self, data):
        colours, radii, numbers = data
        scalars = ndarray_to_vtkarray(colours, radii, numbers)
        self._polydata = vtk.vtkPolyData()
        self._polydata.GetPointData().SetScalars(scalars)
        self.set_coordinates(self._current_frame)

        # self._datamodel.setReader(reader)

    def update_renderer(self):
        """
        Update the renderer
        """
        # deleting old frame
        self.clear_trajectory()

        # creating new polydata
        self._actors = vtk.vtkAssembly()
        for actor in self.create_all_actors():
            self._actors.AddPart(actor)

        # adding polydata to renderer
        self._renderer.AddActor(self._actors)

        # rendering
        if self.reset_camera:
            self._renderer.ResetCamera()
            self.reset_camera = False

        self._iren.Render()


class MolecularViewerWithPicking(MolecularViewer):
    """This class implements a molecular viewer with picking."""

    picked_atoms_changed = Signal(object)

    def __init__(self):
        super().__init__()
        # we set dummy size to something non-zero since we need to be
        # able to see it for picking purposes
        self.dummy_size = 0.1
        self._picking_domain = None
        self._picked_polydata = None
        self._picked_polydata_bonds_exist = False
        self._polydata_opacity = 0.15
        self.picked_atoms = set()
        self.build_events()

    def build_events(self):
        """Build the events."""
        self._iren.AddObserver("LeftButtonPressEvent", self.on_pick)

    def on_pick(self, obj, event=None):
        """Event handler when an atom is mouse-picked with the left mouse button"""

        if not self._reader:
            return

        if self._picking_domain is None:
            return

        picker = vtk.vtkPropPicker()

        picker.AddPickList(self._picking_domain)
        picker.PickFromListOn()

        pos = obj.GetEventPosition()
        picker.Pick(pos[0], pos[1], 0, self._renderer)

        picked_actor = picker.GetActor()
        if picked_actor is None:
            return

        picked_pos = np.array(picker.GetPickPosition())
        coords = self._reader.read_frame(self._current_frame)
        _, idx = KDTree(coords).query(picked_pos)
        self.on_pick_atom(idx)

    def on_pick_atom(self, picked_atom):
        """Change the color of a selected atom"""
        if self._reader is None:
            return

        if picked_atom < 0 or picked_atom >= self._n_atoms:
            return

        if picked_atom in self.picked_atoms:
            self.picked_atoms.remove(picked_atom)
        else:
            self.picked_atoms.add(picked_atom)

        self.update_picked_polydata()
        self.update_renderer()
        self.picked_atoms_changed.emit(self.picked_atoms)

    def change_picked(self, picked: set[int]):
        self.picked_atoms = picked
        self.update_picked_polydata()
        self.update_renderer()

    def update_picked_polydata(self):
        picked = np.array(sorted(list(self.picked_atoms)))
        coords = self._reader.read_frame(self._current_frame)

        atoms = vtk.vtkPoints()
        atoms.SetNumberOfPoints(len(self.picked_atoms))
        for i, j in enumerate(picked):
            x, y, z = coords[j, :]
            atoms.SetPoint(i, x, y, z)
        self._picked_polydata.SetPoints(atoms)

        picked_colours = []
        picked_radii = []
        for i in sorted(list(self.picked_atoms)):
            picked_colours.append(self._colour_manager.colours[i])
            picked_radii.append(self._colour_manager.radii[i])
        scalars = ndarray_to_vtkarray(
            np.array(picked_colours),
            np.array(picked_radii),
            np.arange(len(self.picked_atoms)),
        )
        self._picked_polydata.GetPointData().SetScalars(scalars)

        not_du = np.array(
            [
                i
                for i, j in enumerate(picked)
                if CHEMICAL_ELEMENTS.get_atom_property(
                    self._reader.atom_types[j], "element"
                )
                != "dummy"
            ]
        )

        if self._bonds_visible and len(not_du) >= 1:
            # do not bond atoms to dummy atoms
            rs = coords[picked][not_du]
            covs = np.array(
                [
                    CHEMICAL_ELEMENTS.get_atom_property(
                        self._reader.atom_types[i], "covalent_radius"
                    )
                    for i in picked
                ]
            )[not_du]

            bonds, bonds_exist = self.create_bond_cell_array(rs, covs, not_du)
            if bonds_exist:
                self._picked_polydata.SetLines(bonds)
                self._picked_polydata_bonds_exist = True
                return

        self._picked_polydata_bonds_exist = False

    def reset_all_polydata(self):
        super().reset_all_polydata()
        self._picked_polydata = vtk.vtkPolyData()

    def update_all_polydata(self):
        super().update_all_polydata()
        self.update_picked_polydata()

    def create_all_actors(self):
        actors = []
        if self._polydata is None or self._picked_polydata is None:
            return actors

        line_actor, ball_actor = self.create_traj_actors(
            self._polydata,
            line_opacity=self._polydata_opacity,
            ball_opacity=self._polydata_opacity,
        )
        picked_line_actor, picked_ball_actor = self.create_traj_actors(
            self._picked_polydata
        )
        if self._cell_visible:
            uc_actor = self.create_uc_actor()
            actors.append(uc_actor)
        if self._bonds_visible and self._polydata_bonds_exist:
            actors.append(line_actor)
        if self._bonds_visible and self._picked_polydata_bonds_exist:
            actors.append(picked_line_actor)
        if self._atoms_visible:
            self._picking_domain = ball_actor
            actors.append(ball_actor)
            actors.append(picked_ball_actor)
        else:
            self._picking_domain = None
        return actors

    @Slot(object)
    def take_atom_properties(self, data):
        colours, radii, numbers = data
        scalars = ndarray_to_vtkarray(colours, radii, numbers)
        self._polydata = vtk.vtkPolyData()
        self._polydata.GetPointData().SetScalars(scalars)

        picked_colours = []
        picked_radii = []
        picked_numbers = []
        for i in sorted(list(self.picked_atoms)):
            picked_colours.append(colours[i])
            picked_radii.append(radii[i])
            picked_numbers.append(numbers[i])

        scalars = ndarray_to_vtkarray(
            np.array(picked_colours),
            np.array(picked_radii),
            np.arange(len(self.picked_atoms)),
        )
        self._picked_polydata = vtk.vtkPolyData()
        self._picked_polydata.GetPointData().SetScalars(scalars)

        self.set_coordinates(self._current_frame)



#    This file is part of MDANSE_GUI.
#
#    MDANSE_GUI is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program.  If not, see <https://www.gnu.org/licenses/>.
#
from qtpy.QtCore import Slot, Signal
from qtpy.QtWidgets import QTextBrowser
from qtpy.QtCore import Signal, Slot, Qt
from qtpy.QtGui import QStandardItem

from MDANSE.MLogging import LOG
from MDANSE.Framework.QVectors.IQVectors import IQVectors
from MDANSE.Framework.Units import measure
from MDANSE_GUI.Widgets.ResolutionWidget import ResolutionCalculator, widget_text_map


class SimpleInstrument:

    sample_options = ["isotropic", "crystal"]
    technique_options = ["QENS", "INS"]
    resolution_options = [str(x) for x in widget_text_map.keys()]
    qvector_options = [str(x) for x in IQVectors.indirect_subclasses()]
    energy_units = ["meV", "1/cm", "THz"]
    momentum_units = ["1/ang", "1/nm", "1/Bohr"]

    def __init__(self, optional_qitem_reference: QStandardItem = None) -> None:
        self._list_item = optional_qitem_reference
        self._name = "Generic neutron instrument"
        self._name_is_fixed = False
        self._sample = "isotropic"
        self._technique = "QENS"
        self._resolution_type = "Gaussian"
        self._resolution_fwhm = 0.1
        self._resolution_unit = "meV"
        self._qvector_type = "SphericalQVectors"
        self._q_min = 0.1
        self._q_max = 10.0
        self._q_unit = "1/ang"
        self._q_step = 1.0
        self._q_width = 1.0
        self._axis_1 = [1, 0, 0]
        self._axis_2 = [0, 1, 0]
        self._vectors_per_shell = 100
        self._configured = False
        self._model_index = -1

    @classmethod
    def inputs(cls):
        input_list = [
            ["_name", "QLineEdit", "str"],
            ["_sample", "QComboBox", "sample_options"],
            ["_technique", "QComboBox", "technique_options"],
            ["_resolution_type", "QComboBox", "resolution_options"],
            ["_resolution_fwhm", "QLineEdit", "float"],
            ["_resolution_unit", "QComboBox", "energy_units"],
            ["_qvector_type", "QComboBox", "qvector_options"],
            ["_q_unit", "QComboBox", "momentum_units"],
            ["_q_min", "QLineEdit", "float"],
            ["_q_max", "QLineEdit", "float"],
            ["_q_step", "QLineEdit", "float"],
            ["_q_width", "QLineEdit", "float"],
            ["_vectors_per_shell", "QLineEdit", "int"],
            ["_axis_1", "VectorWidget", "float"],
            ["_axis_2", "VectorWidget", "float"],
        ]
        return input_list

    def update_item(self):
        if self._list_item is None:
            return
        self._list_item.setData(self._name, role=Qt.ItemDataRole.DisplayRole)
        self._list_item.setText(self._name)

    def create_resolution_params(self):
        if not self._configured:
            return
        general_resolution = ("gaussian", {"mu": 0.0, "sigma": 0.2})
        calculator = ResolutionCalculator()
        try:
            calculator.update_model(self._resolution_type)
        except Exception as e:
            LOG.error(f"update_model failed: {e}")
        try:
            calculator.recalculate_peak(
                self._resolution_fwhm, 0.0, 0.0, self._resolution_unit
            )
        except Exception as e:
            LOG.error(f"recalculate_peak failed: {e}")
        _, results = calculator.summarise_results()
        results.pop("function", None)
        mdanse_tuple = (widget_text_map[self._resolution_type], results)
        self._resolution_results = mdanse_tuple
        return mdanse_tuple

    def sanitize_numbers(self):
        results = []
        for entry in [self._q_step, self._q_min, self._q_max, self._q_width]:
            try:
                new_entry = float(entry)
            except ValueError:
                if "e" in entry:
                    try:
                        new_entry = float(entry.split("e")[0])
                    except ValueError:
                        new_entry = 1.0
            results.append(new_entry)
        return results

    def create_q_vector_params(self):
        if not self._configured:
            return
        cov_type = self._qvector_type
        try:
            qvec_generator = IQVectors.create(cov_type, None)
        except ValueError:
            return ("No qvectors", {})
        except AttributeError:
            return (cov_type, {})
        qvec_generator.build_configuration()
        param_dictionary = {}
        _q_step, _q_min, _q_max, _q_width = self.sanitize_numbers()
        try:
            conversion_factor = measure(1.0, iunit=self._q_unit).toval("1/nm")
        except:
            raise ValueError(f"Could not convert unit: {self._q_unit}")
        else:
            conversion_factor = float(conversion_factor)
        if "shells" in qvec_generator._configuration:
            q_step = round(conversion_factor * float(_q_step), 6)
            q_min = round(conversion_factor * float(_q_min), 6)
            q_max = round(conversion_factor * float(_q_max), 6)
            param_dictionary["shells"] = [q_min, q_max, q_step]
        if "width" in qvec_generator._configuration:
            width = round(conversion_factor * float(_q_width), 6)
            param_dictionary["width"] = width
        if "n_vectors" in qvec_generator._configuration:
            num_vectors = int(self._vectors_per_shell)
            param_dictionary["n_vectors"] = num_vectors
        if "axis" in qvec_generator._configuration:
            ax_vector = [float(x) for x in self._axis_1.split(",")]
            param_dictionary["axis"] = ax_vector
        if "axis_1" in qvec_generator._configuration:
            ax_vector = [float(x) for x in self._axis_1.split(",")]
            param_dictionary["axis_1"] = ax_vector
        if "axis_2" in qvec_generator._configuration:
            ax_vector = [float(x) for x in self._axis_2.split(",")]
            param_dictionary["axis_2"] = ax_vector
        mdanse_tuple = (
            cov_type,
            param_dictionary,
        )
        return mdanse_tuple

    def filter_qvector_generator(self):
        new_list = [str(x) for x in self.qvector_options]
        if self._sample == "isotropic":
            new_list = [str(x) for x in new_list if not "Lattice" in x]
        if self._technique == "QENS":
            new_list = [str(x) for x in new_list if "Spherical" in x]
        return new_list

    def pick_qvector_generator(self) -> str:
        qgenerator = "SphericalQVectors"
        if self._sample == "isotropic" and self._technique == "QENS":
            qgenerator = "SphericalQVectors"
        return qgenerator


class InstrumentInfo(QTextBrowser):
    error = Signal(str)

    def __init__(self, *args, **kwargs):
        self._header = kwargs.pop("header", "")
        self._footer = kwargs.pop("footer", "")
        super().__init__(*args, **kwargs)
        self.setOpenExternalLinks(True)
        self.setHtml("")

    @Slot(object)
    def update_panel(self, incoming: SimpleInstrument):
        if incoming is None:
            return
        try:
            filtered = self.filter(incoming)
        except Exception as e:
            LOG.error(f"Error in InstrumentInfo: {e}")
            self.setHtml("")
        else:
            self.setHtml(filtered)

    @Slot(str)
    def append_text(self, new_text: str):
        self.append(new_text)

    def filter(self, instrument_object: SimpleInstrument, line_break="<br />") -> str:
        """Extracts and displays results calculated by SimpleInstrument.

        Parameters
        ----------
        instrument_object : SimpleInstrument
            Class containing a neutron instrument definition
        line_break : str, optional
            Text formatting option, by default "<br />" (HTML line break)

        Returns
        -------
        str
            Formatted text
        """

        new_text = ""
        if self._header:
            new_text += self._header + line_break
        new_text += "<b>" + instrument_object._name + "</b>" + line_break + line_break
        new_text += "MDANSE resolution input" + line_break
        new_text += str(instrument_object.create_resolution_params()) + line_break
        new_text += "MDANSE q-vector generator input" + line_break
        new_text += str(instrument_object.create_q_vector_params()) + line_break
        if self._footer:
            new_text += line_break + self._footer
        return new_text



#    This file is part of MDANSE_GUI.
#
#    MDANSE_GUI is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program.  If not, see <https://www.gnu.org/licenses/>.
#
from qtpy.QtCore import Slot, Signal
from qtpy.QtWidgets import QTextBrowser

from MDANSE.Framework.Jobs.IJob import IJob


class AnalysisInfo(QTextBrowser):
    error = Signal(str)

    def __init__(self, *args, **kwargs):
        self._header = kwargs.pop("header", "")
        self._footer = kwargs.pop("footer", "")
        super().__init__(*args, **kwargs)
        self.setOpenExternalLinks(True)

    @Slot(object)
    def update_panel(self, incoming: object):
        filtered = self.filter(incoming)
        self.setHtml(filtered)

    def filter(self, some_text: str, line_break="<br />"):
        new_text = ""
        if self._header:
            new_text += self._header + line_break
        if some_text is not None:
            new_text += line_break.join([x.strip() for x in some_text.split("\n")])
        if self._footer:
            new_text += line_break + self._footer
        return new_text

    def summarise_chemical_system(self, job_name):
        try:
            temp_instance = IJob.create(job_name)
        except:
            return ""
        text = "\n ==== Input Parameter summary ==== \n"
        params = temp_instance.get_default_parameters()
        for key, value in params.items():
            try:
                text += f"parameters['{key}'] = {value[0]} # {value[1]} \n"
            except:
                continue
        text += " ===== \n"
        return text



#    This file is part of MDANSE_GUI.
#
#    MDANSE_GUI is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program.  If not, see <https://www.gnu.org/licenses/>.
#
from qtpy.QtWidgets import QLineEdit

from MDANSE_GUI.InputWidgets.WidgetBase import WidgetBase


class BackupWidget(WidgetBase):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        source_object = kwargs.get("source_object", None)
        self._field = QLineEdit(str(self._configurator.default))
        self._field.setPlaceholderText(str(self._configurator.default))
        self._layout.addWidget(self._field)
        if self._tooltip:
            tooltip_text = self._tooltip
        else:
            tooltip_text = "There should have been a specialised widget for this parameter, but there isn't."
        self._field.textChanged.connect(self.updateValue)
        self._field.setToolTip(tooltip_text)
        self.updateValue()

    def get_widget_value(self):
        """Collect the results from the input widgets and return the value."""
        temp_text = self._field.text().strip()
        if temp_text == "None" or temp_text == "":
            result = str(self._configurator.default)
            self._empty = True
        else:
            result = temp_text
            self._empty = False
        return result



#    This file is part of MDANSE_GUI.
#
#    MDANSE_GUI is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program.  If not, see <https://www.gnu.org/licenses/>.
#

from qtpy.QtWidgets import QLineEdit, QRadioButton, QButtonGroup, QLabel
from qtpy.QtCore import Slot

from MDANSE_GUI.InputWidgets.WidgetBase import WidgetBase


class ProjectionWidget(WidgetBase):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        source_object = kwargs.get("source_object", None)
        bgroup = QButtonGroup(self._base)
        for id, blabel in enumerate(["None", "Axial", "Planar"]):
            rbutton = QRadioButton(blabel, parent=self._base)
            bgroup.addButton(rbutton, id=id)
            self._layout.addWidget(rbutton)
            if id == 0:
                rbutton.setChecked(True)
        self._changing_label = QLabel("N/A", parent=self._base)
        self._layout.addWidget(self._changing_label)
        vfields = []
        for _ in range(3):
            temp = QLineEdit("0", self._base)
            self._layout.addWidget(temp)
            vfields.append(temp)
        self._button_group = bgroup
        self._vector_fields = vfields
        self._mode = 0
        self._button_group.idClicked.connect(self.button_switched)
        self.updateValue()
        if self._tooltip:
            tooltip_text = self._tooltip
        else:
            tooltip_text = "The spatial properties in the analysis can be projected on a plane or on an axis, as chosen here."
        for wid in vfields:
            wid.setToolTip(tooltip_text)
        for wid in self._button_group.buttons():
            wid.setToolTip(tooltip_text)
        self.button_switched(0)
        self._button_group.buttonClicked.connect(self.updateValue)
        for vfield in self._vector_fields:
            vfield.textChanged.connect(self.updateValue)

    def configure_using_default(self):
        """This is too complex to have a default value"""

    @Slot(int)
    def button_switched(self, button_number: int):
        button_id = self._button_group.checkedId()
        if button_id == 0:
            self._changing_label.setText("N/A")
            for field in self._vector_fields:
                field.setEnabled(False)
        elif button_id == 1:
            self._changing_label.setText("Axis direction")
            for field in self._vector_fields:
                field.setEnabled(True)
        else:
            self._changing_label.setText("Plane normal")
            for field in self._vector_fields:
                field.setEnabled(True)
        self._mode = button_id

    def get_widget_value(self):
        """Collect the results from the input widgets and return the value."""
        if self._mode == 0:
            return ("NullProjector", [])
        vector = [x.text() for x in self._vector_fields]
        if self._mode == 1:
            return ("AxialProjector", vector)
        else:
            return ("PlanarProjector", vector)



#    This file is part of MDANSE_GUI.
#
#    MDANSE_GUI is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program.  If not, see <https://www.gnu.org/licenses/>.
#
import os
from typing import Tuple
from multiprocessing import Queue
from multiprocessing.connection import Connection
from multiprocessing.synchronize import Event

from qtpy.QtCore import QObject, Slot, Signal

from MDANSE.MLogging import LOG
from MDANSE.Framework.Status import Status


class JobCommunicator(QObject):
    target = Signal(int)
    progress = Signal(int)
    finished = Signal(bool)
    oscillate = Signal()

    def status_update(self, input: Tuple):
        key, value = input
        if key == "FINISHED":
            self.finished.emit(value)
            self.terminate_the_process()
        elif key == "STEP":
            self.progress.emit(value)
        elif key == "STARTED":
            if value is not None:
                self.target.emit(value)
            else:
                self.oscillate.emit()
        elif key == "COMMUNICATION":
            LOG.info(f"Communication with the subprocess is now {value}")
            self.finished.emit(value)
            self.terminate_the_process()

    @Slot()
    def terminate_the_process(self):
        LOG.info(f"JobCommunicator PID: {os.getpid()} started 'terminate_the_process")
        try:
            self._process.terminate()
        except:
            return
        else:
            try:
                self._process.close()
            except:
                return


class JobStatusProcess(Status):
    def __init__(
        self,
        pipe: "Connection",
        queue_0: Queue,
        queue_1: Queue,
        pause_event: "Event",
        **kwargs,
    ):
        super().__init__()
        self._pipe = pipe
        self._queue_0 = queue_0
        self._queue_1 = queue_1
        self._state = {}  # for compatibility with JobStatus
        self._progress_meter = 0
        self._pause_event = pause_event
        self._pause_event.set()

    @property
    def state(self):
        return self._state

    def finish_status(self):
        self._pipe.send(("FINISHED", True))

    def start_status(self):
        LOG.info(f"JobStatusProcess PID: {os.getpid()} started 'start_status")
        try:
            temp = int(self._nSteps)
        except:
            self._pipe.send(("STARTED", None))
        else:
            self._pipe.send(("STARTED", temp))
        # self._updateStep = 1

    def stop_status(self):
        self._pipe.send(("FINISHED", False))

    def update_status(self):
        self._progress_meter += 1
        temp = int(self._progress_meter) * self._updateStep
        self._pipe.send(("STEP", temp))

    def fixed_status(self, current_progress: int):
        temp = int(current_progress) * self._updateStep
        self._pipe.send(("STEP", temp))



#    This file is part of MDANSE_GUI.
#
#    MDANSE_GUI is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program.  If not, see <https://www.gnu.org/licenses/>.
#
from functools import partial
from qtpy.QtCore import Slot
from qtpy.QtWidgets import QWidget

from MDANSE.Framework.Converters.Converter import Converter

from MDANSE_GUI.Tabs.GeneralTab import GeneralTab
from MDANSE_GUI.Tabs.Layouts.MultiPanel import MultiPanel
from MDANSE_GUI.Session.LocalSession import LocalSession
from MDANSE_GUI.Tabs.Visualisers.TextInfo import TextInfo
from MDANSE_GUI.Tabs.Visualisers.Action import Action
from MDANSE_GUI.Tabs.Models.JobTree import JobTree
from MDANSE_GUI.Tabs.Views.ActionsTree import ActionsTree


tab_label = """Convert your trajectory to the MDANSE MDT format.
If you cannot find a dedciated converter
for your MD engine, the ASE converter
can be tried as a backup option.
"""


class ConverterTab(GeneralTab):
    """The tab for choosing and starting a new job."""

    def __init__(self, *args, **kwargs):
        self.action = kwargs.pop("action")
        super().__init__(*args, **kwargs)
        self._current_trajectory = ""
        self._job_starter = None
        self.action._parent_tab = self
        self._visualiser._parent_tab = self

    def set_job_starter(self, job_starter):
        self._job_starter = job_starter
        self.action.new_thread_objects.connect(self._job_starter.startProcess)
        self.action.run_and_load.connect(self._job_starter.startProcessAndLoad)

    @Slot(str)
    def set_current_trajectory(self, new_name: str):
        self._current_trajectory = new_name

    def grouped_settings(self):
        results = super().grouped_settings()
        results += [
            [
                "Execution",
                {"auto-load": "True"},
                {
                    "auto-load": "Unless manually switched off, the GUI will try to load the job results when the job is finished."
                },
            ]
        ]
        return results

    @classmethod
    def standard_instance(cls):
        action = Action()
        the_tab = cls(
            window,
            name="AvailableJobs",
            model=JobTree(parent_class=Converter),
            view=ActionsTree(),
            visualiser=action,
            layout=partial(
                MultiPanel,
                left_panels=[
                    TextInfo(
                        header="MDANSE Converter",
                        footer="Look up our Read The Docs page:"
                        + "https://mdanse.readthedocs.io/en/protos/",
                    )
                ],
            ),
            label_text=tab_label,
            action=action,
        )
        return the_tab

    @classmethod
    def gui_instance(
        cls,
        parent: QWidget,
        name: str,
        session: LocalSession,
        settings,
        logger,
        **kwargs,
    ):
        action = Action()
        the_tab = cls(
            parent,
            name=name,
            session=session,
            settings=settings,
            logger=logger,
            model=kwargs.get("model", JobTree(parent_class=Converter)),
            view=ActionsTree(),
            visualiser=action,
            layout=partial(
                MultiPanel,
                left_panels=[
                    TextInfo(
                        header="MDANSE Converter",
                        footer="Look up our "
                        + '<a href="https://mdanse.readthedocs.io/en/protos/">Read The Docs</a>'
                        + " page.",
                    )
                ],
            ),
            label_text=tab_label,
            action=action,
        )
        action.set_settings(the_tab._settings)
        return the_tab


if __name__ == "__main__":
    import sys
    from qtpy.QtWidgets import QApplication, QMainWindow

    app = QApplication(sys.argv)
    window = QMainWindow()
    action = Action()
    the_tab = ConverterTab(
        window,
        name="AvailableJobs",
        model=JobTree(parent_class=Converter),
        view=ActionsTree(),
        visualiser=action,
        layout=partial(
            MultiPanel,
            left_panels=[
                TextInfo(
                    header="MDANSE Converters",
                    footer="Look up our "
                    + '<a href="https://mdanse.readthedocs.io/en/protos/">Read The Docs</a>'
                    + " page.",
                )
            ],
        ),
        label_text=tab_label,
        action=action,
    )
    the_tab._current_trajectory = "/Users/maciej.bartkowiak/an_example/BLAH.mdt"
    window.setCentralWidget(the_tab._core)
    window.show()
    app.exec()



#    This file is part of MDANSE_GUI.
#
#    MDANSE_GUI is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program.  If not, see <https://www.gnu.org/licenses/>.
#
import pytest
import tempfile
import os

import numpy as np

from MDANSE.Chemistry.ChemicalEntity import Atom, ChemicalSystem
from MDANSE.MolecularDynamics.Configuration import RealConfiguration
from MDANSE.MolecularDynamics.Trajectory import Trajectory, TrajectoryWriter
from MDANSE.Framework.InputData.HDFTrajectoryInputData import HDFTrajectoryInputData

from MDANSE_GUI.MolecularViewer.readers.hdf5wrapper import HDF5Wrapper

N_ATOMS = 4
N_TIMESTEPS = 150


@pytest.fixture(scope="module")
def chemical_system():
    temp = ChemicalSystem("Dummy test system")
    nAtoms = N_ATOMS
    for i in range(nAtoms):
        temp.add_chemical_entity(Atom(symbol="H"))
    return temp


@pytest.fixture(scope="module")
def sample_configuration(chemical_system):
    unit_cell = 15.0 * np.eye(3)
    coords = np.empty((N_ATOMS, 3), dtype=float)
    coords[0] = [1.0, 1.0, 1.0]
    coords[1] = [1.0, 2.0, 1.0]
    coords[2] = [10.0, 1.0, 5.11]
    coords[3] = [10.0, 2.0, 5.09]
    temp = RealConfiguration(
        chemical_system,
        coords,
        # unit_cell
    )
    return temp


@pytest.fixture(scope="module")
def sample_trajectory(chemical_system, sample_configuration):
    # here we write to a file
    fdesc, fname = tempfile.mkstemp()
    os.close(fdesc)
    writer = TrajectoryWriter(fname, chemical_system, n_steps=N_TIMESTEPS)
    for n, ts in enumerate(np.arange(N_TIMESTEPS)):
        writer.chemical_system.configuration = sample_configuration
        writer.dump_configuration(ts)
    return fname


@pytest.fixture(scope="module")
def wrapper(sample_trajectory):
    traj_object = HDFTrajectoryInputData(sample_trajectory)
    wrapper = HDF5Wrapper(
        sample_trajectory, traj_object.trajectory, traj_object.chemical_system
    )
    return wrapper


def test_frame_read(wrapper: HDF5Wrapper):
    coords = wrapper.read_frame(0)
    assert len(coords.shape) == 2



#    This file is part of MDANSE_GUI.
#
#    MDANSE_GUI is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program.  If not, see <https://www.gnu.org/licenses/>.
#
import glob
import itertools
import os
import os.path

from qtpy.QtWidgets import QLineEdit, QPushButton, QFileDialog, QLabel, QComboBox
from qtpy.QtCore import Slot, Qt

from MDANSE.Framework.Configurators.OutputFilesConfigurator import (
    OutputFilesConfigurator,
)
from MDANSE.MLogging import LOG

from MDANSE_GUI.InputWidgets.WidgetBase import WidgetBase
from .CheckableComboBox import CheckableComboBox


class OutputFilesWidget(WidgetBase):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, layout_type="QGridLayout", **kwargs)
        default_value = self._configurator.default
        try:
            parent = kwargs.get("parent", None)
            self.default_path = parent.default_path
        except KeyError:
            self.default_path = "."
            LOG.error("KeyError in OutputFilesWidget - can't get default path.")
        except AttributeError:
            self.default_path = "."
            LOG.error("AttributeError in OutputFilesWidget - can't get default path.")
        self.file_association = "Output file name (*)"
        self._value = default_value
        self._field = QLineEdit(default_value[0], self._base)
        self._field.setPlaceholderText(default_value[0])
        self.type_box = CheckableComboBox(self._base)
        self.type_box.addItems(self._configurator.formats)
        self.type_box.set_default("MDAFormat")
        # self.type_box.setCurrentText(default_value[1])
        browse_button = QPushButton("Browse", self._base)
        browse_button.clicked.connect(self.file_dialog)
        label = QLabel("Log file output:")
        label.setAlignment(Qt.AlignRight | Qt.AlignVCenter)
        self.logs_combo = QComboBox(self._base)
        self.logs_combo.addItems(OutputFilesConfigurator.log_options)
        self._layout.addWidget(self._field, 0, 0)
        self._layout.addWidget(self.type_box, 0, 1)
        self._layout.addWidget(browse_button, 0, 2)
        self._layout.addWidget(label, 1, 0)
        self._layout.addWidget(self.logs_combo, 1, 1)
        self._default_value = default_value
        self._field.textChanged.connect(self.updateValue)
        self.type_box.lineEdit().textChanged.connect(self.updateValue)
        self.default_labels()
        self.update_labels()
        self.updateValue()
        if self._tooltip:
            tooltip_text = self._tooltip
        else:
            tooltip_text = "Analysis output will be saved under this name, and using the selected file types"
        self._field.setToolTip(tooltip_text)
        self.type_box.setToolTip(tooltip_text)

    def default_labels(self):
        """Each Widget should have a default tooltip and label,
        which will be set in this method, unless specific
        values are provided in the settings of the job that
        is being configured."""
        if self._label_text == "":
            self._label_text = "OutputFilesWidget"
        if self._tooltip == "":
            self._tooltip = "Analysis output will be saved under this name, and using the selected file types"

    @Slot()
    def file_dialog(self):
        """A Slot defined to allow the GUI to be updated based on
        the new path received from a FileDialog.
        This will start a FileDialog, take the resulting path,
        and emit a signal to update the value show by the GUI.
        """
        new_value = QFileDialog.getSaveFileName(
            self._base,  # the parent of the dialog
            "Save files",  # the label of the window
            self.default_path,  # the initial search path
            self.file_association,  # text string specifying the file name filter.
        )
        if len(new_value[0]) > 0:
            self._field.setText(new_value[0])
            self.updateValue()

    @staticmethod
    def _get_unique_filename(directory, basename):
        filesInDirectory = [
            os.path.join(directory, e)
            for e in itertools.chain(
                glob.iglob(os.path.join(directory, "*")),
                glob.iglob(os.path.join(directory, ".*")),
            )
            if os.path.isfile(os.path.join(directory, e))
        ]
        basenames = [os.path.splitext(f)[0] for f in filesInDirectory]

        initialPath = path = os.path.join(directory, basename)
        comp = 1
        while True:
            if path in basenames:
                path = "%s(%d)" % (initialPath, comp)
                comp += 1
                continue
            return path

    def get_widget_value(self):
        filename = self._field.text()
        if len(filename) < 1:
            filename = self._default_value[0]

        formats = self.type_box.checked_values()
        log_level = self.logs_combo.currentText()

        return (filename, formats, log_level)

    def set_data(self, datakey):
        basename = "%s_%s" % (
            os.path.splitext(os.path.basename(datakey))[0],
            self._parent.type,
        )
        trajectoryDir = os.path.dirname(datakey)

        path = OutputFilesWidget._get_unique_filename(trajectoryDir, basename)

        self._filename.SetValue(path)



#    This file is part of MDANSE_GUI.
#
#    MDANSE_GUI is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program.  If not, see <https://www.gnu.org/licenses/>.
#

import numpy as np


class PyConnectivity:
    """At the moment MDANSE_GUI does not contain any checks for connectivity.
    The intended solution is to include a connectivity scanner in the
    trajectory converter.
    In the meantime, since the MolecularViewer expects a PyConnectivity object,
    this dummy has been created. It never finds any bonds.
    """

    def __init__(self, *args, **kwargs):
        pass

    def add_point(self, index: int, point: np.ndarray, radius: float) -> bool:
        return True

    def find_collisions(self, tolerance: float) -> dict:
        return {}

    def get_neighbour(self, point: np.ndarray) -> int:
        return 0



#    This file is part of MDANSE_GUI.
#
#    MDANSE_GUI is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program.  If not, see <https://www.gnu.org/licenses/>.
#
from math import floor
from .RangeWidget import RangeWidget


class DistHistCutoffWidget(RangeWidget):

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def setup_fields(self, *args, **kwargs):
        start = 0.0
        end = floor(self._configurator.get_largest_cutoff() * 100) / 100
        step = 0.01
        super().setup_fields(*args, default=(start, end, step), **kwargs)
